#include <stdlib.h>
#include <stdio.h>
#include "container.h"

struct container {
  int count, size, start;
  void** a;
};

container* make_container() {
  container* ac;

  ac = (container*)malloc(sizeof(container));
  ac->count = 0;
  ac->start = 0;
  ac->size = 1;
  ac->a = (void**)malloc(ac->size * sizeof(void*));

  return ac;
}

static void copy_array(void** to, void** from, int len) {
  int i;
  for (i = len; i-- > 0; ) {
    to[i] = from[i];
  }
}

static void maybe_grow_array(container* ac) {
  void **new_a;
  if (ac->count >= ac->size) {
    new_a = (void**)malloc(ac->size * 2 * sizeof(void*));
    if (ac->start + ac->count < ac->size) {
      copy_array(new_a, ac->a + ac->start, ac->count);
    } else {
      int c1 = ac->size - ac->start;
      copy_array(new_a, ac->a + ac->start, c1);
      copy_array(new_a + c1, ac->a, ac->count - c1);
    }
    ac->size *= 2;
    ac->start = 0;
    ac->a = new_a;
  }
}

void add_to_front(container* ac, void* i) {
  maybe_grow_array(ac);

  --ac->start;
  if (ac->start < 0)
    ac->start = ac->start + ac->size;
  ac->a[ac->start] = i;
  ac->count++;
}

void add_to_back(container* ac, void* i) {
  maybe_grow_array(ac);
  
  ac->a[(ac->start + ac->count) % ac->size] = i;
  ac->count++;
}

void* remove_from_front(container* ac) {
  void* i;

  i = ac->a[ac->start];
  ac->start++;
  if (ac->start >= ac->size)
    ac->start = ac->start - ac->size;
  --ac->count;

  return i;
}

void* remove_from_back(container* ac) {
  void* i;

  i = ac->a[(ac->start + ac->count - 1) % ac->size];
  --ac->count;

  return i;
}

int get_count(container* ac) {
  return ac->count;
}
