#include <stdio.h>
#include <stdlib.h>
#include "container.h"

/* Link this program with "container.c" */

/****************************************/

struct int_node {
  int val;
  struct int_node *left;
  struct int_node *right;
};

typedef struct int_node * tree;

tree leaf(int n) {
  tree t = (tree)malloc(sizeof(struct int_node));

  t->val = n;
  t->left = NULL;
  t->right = NULL;

  return t;
}

tree branch(int n, tree l, tree r) {
  tree t = (tree)malloc(sizeof(struct int_node));

  t->val = n;
  t->left = l;
  t->right = r;

  return t;
}

void print_tree(tree t, int indent) {
  if (t != NULL) {
    int i;

    print_tree(t->right, indent + 2);

    for (i = 0; i < indent; i++)
      printf(" ");
    printf("- %d\n", t->val);

    print_tree(t->left, indent + 2);
  }
}

/****************************************/

void print_preorder(tree t)
{
  container* c = make_container();
  
  if (t != NULL)
    add_to_back(c, t);
  while (get_count(c) > 0) {
    t = remove_from_back(c);
    printf(" %d\n", t->val);
    if (t->right != NULL)
      add_to_back(c, t->right);
    if (t->left != NULL)
      add_to_back(c, t->left);
  }
}

void print_breadth_first(tree t)
{
  container* c = make_container();
 
  if (t != NULL)
    add_to_back(c, t);
  while (get_count(c) > 0) {
    t = remove_from_front(c);
    printf(" %d\n", t->val);
    if (t->left != NULL)
      add_to_back(c, t->left);
    if (t->right != NULL)
      add_to_back(c, t->right);
  }
}

typedef struct todo {
  tree t;
  int what; /* 0 = explore, 1 = print */
} todo;

todo* make_todo(tree t, int what) {
  todo* td;
  td = (todo*)malloc(sizeof(todo));
  td->t = t;
  td->what = what;
  return td;
}

void original_print_postordered(tree t)
{
  container* c = make_container();
  todo *td;
  
  if (t != NULL)
    add_to_back(c, make_todo(t, 0));
  while (get_count(c) > 0) {
    td = remove_from_back(c);
    t = td->t;
    if (td->what == 0) {
      add_to_back(c, make_todo(t, 1));
      if (t->right != NULL)
        add_to_back(c, make_todo(t->right, 0));
      if (t->left != NULL)
        add_to_back(c, make_todo(t->left, 0));
    } else {
      printf(" %d\n", t->val);
    }
  }
}

void original_print_inordered(tree t)
{
  container* c = make_container();
  todo *td;
  
  if (t != NULL)
    add_to_back(c, make_todo(t, 0));
  while (get_count(c) > 0) {
    td = remove_from_back(c);
    t = td->t;
    if (td->what == 0) {
      if (t->right != NULL)
        add_to_back(c, make_todo(t->right, 0));
      add_to_back(c, make_todo(t, 1));
      if (t->left != NULL)
        add_to_back(c, make_todo(t->left, 0));
    } else {
      printf(" %d\n", t->val);
    }
  }
}

void print_ordered(tree t, int after)
{
  container* c = make_container();
  todo *td;
  
  if (t != NULL)
    add_to_back(c, make_todo(t, 0));
  while (get_count(c) > 0) {
    td = remove_from_back(c);
    t = td->t;
    if (td->what == 0) {
      if (after)
        add_to_back(c, make_todo(t, 1));
      if (t->right != NULL)
        add_to_back(c, make_todo(t->right, 0));
      if (!after)
        add_to_back(c, make_todo(t, 1));
      if (t->left != NULL)
        add_to_back(c, make_todo(t->left, 0));
    } else {
      printf(" %d\n", t->val);
    }
  }
}

void print_postorder(tree t)
{
  print_ordered(t, 1);
}

void print_inorder(tree t)
{
  print_ordered(t, 0);
}

/****************************************/

int main() {
  tree t = branch(5,
                  branch(3,
                         branch(1, NULL, leaf(2)),
                         leaf(4)),
                  branch(8,
                         branch(7, leaf(6), NULL),
                         leaf(9)));

  print_tree(t, 0);

  printf("Preorder:\n");
  print_preorder(t);

  printf("Breadth-first:\n");
  print_breadth_first(t);

  printf("Postorder:\n");
  print_postorder(t);

  printf("Inorder:\n");
  print_inorder(t);

  return 0;
}
